#!/usr/bin/env python

"""Tests for `pydeflate` package."""

import pytest
import pandas as pd


from pydeflate.deflate import deflate


def dummy_data():
    
    """Data for South Africa taken from World Bank"""
    
    years = list(range(2008,2020))
    gdp_current = [286.77, 295.936, 375.349, 416.419, 396.333,
                   366.829, 350.905, 317.621, 296.357, 349.544,
                   368.289, 351.432]
    
    gdp_constant = [369.967, 364.276, 375.349, 387.677, 396.257,
                    406.105, 413.606, 418.543, 420.213, 426.157,
                    429.511, 430.167]

    df = pd.DataFrame(list(zip(years, gdp_current, gdp_constant)), columns=['year',
                                                                        'gdp_current',
                                                                        'gdp_constant'])

    df['iso_code'] = 'ZAF'
    
    return df

def test_deflate_data():
    
    df = dummy_data()
    
    df2c = deflate.deflate_data(df, method='gdp_ratio', base=2010,
                                value_column='gdp_current')
    
    df2c['difference'] = abs(100*((df2c.gdp_constant - df2c.value_constant)/df2c.gdp_constant))
    
    assert round(df2c.difference.mean(),1) <0.1
    
  


"""Top-level package for pydeflate."""

__author__ = """Jorge Rivera"""
__email__ = 'jorge.rivera@one.org'
__version__ = '0.1.3'


from pydeflate.deflate.deflate import deflate_data

from pydeflate.utils import update_data



#!/usr/bin/env python

"""Tests for `pydeflate` package."""

import pytest


from pydeflate import utils


def test_update_gdp():
    # download the data using WB API
    constant, current = utils.update_gdp(test=True)

    if (constant is None) or (current is None):
        raise AssertionError('Could not use WBGAPI to download data')

    # check if the base year continues to be 2010, and if data downloaded correctly
    constant = constant.loc[(constant.year == 2010) &
                            (constant.iso_code == 'USA')]
    current = current.loc[(current.year == 2010) & (current.iso_code == 'USA')]

    result = int(current.value/constant.value)

    assert result == 1


def test_update_exchange():
    exchange = utils.update_exchange(test=True)

    if exchange is None:
        raise AssertionError('Could not use WBGAPI to download data')

    assert int(exchange.loc[(exchange.year == 2010) &
                            (exchange.iso_code == 'USA')].value) == 1


def test_update_data():
    ...


def test_clean_df():
    ...


def test_exchange_rates():
    ...


def test_country_rebase():
    ...


def test_gdp_deflator():
    ...


def test_exchange_deflator():
    ...


def test_gdp_xe_deflator():
    ...


def test_gdp_factor():
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 21 13:06:41 2021

@author: jorge
"""

# %% Import required packages and create paths
import pydeflate.utils as utils
import pandas as pd

# %%


def deflate_data(df: pd.DataFrame, method: str = 'gdp_xe', country_column: str = 'iso_code', 
                 year_column: str = 'year', value_column: str = 'value', 
                 base: int = 2019, to_current: bool = False) -> pd.DataFrame:
    """
    Converts a column in a pandas dataframe to constant prices or vice-versa. 
    Two methods are posible at present: 
        - 'gdp_xe', which creates a common unit of measurement, taking into account changes
        in prices (GDP deflators) and changes in exchange rates over time, for each
        country. Data passed must be passed in current prices unless "to_current" is true.

        - 'gdp_ratio', which uses a GDP ratio approach, where a factor is applied to 
        the LCU data to transform it to constant prices. This factor is obtained by dividing
        GDP LCU current by GDP USD Constant. Data passed must be passed in current prices
        unless "to_current" is true.



    Parameters
    ----------
    df : pd.DataFrame
        A DataFrame, tidy-long, with iso3 codes to identify countries, and years as int or datetime.
    method : str, optional
        Currently two methodologies are accepted: 'gdp_xe' or 'gdp_ratio'. The default is 'gdp_xe'.
    country_column : str, optional
        The column containing iso3 codes. The default is iso_code.
    year_column : str, optional
        The column containing the years. The default is year.  
    value_column : str, optional
        The column containing the values to deflate. The default is value.
    base : int, optional
        The desired base year for the data. The default is 2019.
    to_current : bool, optional
        Whether to convert from constant to current instead of current to constant. The default is False.


    Returns
    -------
    Pandas DataFrame
        The original dataframe including an additional column called either 'value_current' or 
        'value_constant' depending on the type of operation performed.

    """

    # Error handling
    if country_column not in df.columns:
        raise ValueError(
            f'{country_column} not a valid column name for countries')

    if year_column not in df.columns:
        raise ValueError(f'{year_column} not a valid column name for years')

    if value_column not in df.columns:
        raise ValueError(f'{value_column} not a valid column name for values')

    if df[year_column].dtype not in [int, 'datetime64[ns]']:
        raise TypeError(
            'Years are not correctly formatted as integers or datetime objects')

    if df[value_column].dtype not in [int, float]:
        raise TypeError('Values must be integers or floats')

    if method not in ['gdp_xe', 'gdp_ratio']:
        raise TypeError(
            'A valid method must be selected: "gdp_xe" or "gdp_ratio".')

    # Convert year column to datetime if needed
    if df[year_column].dtype == int:
        date_int = True
        df[year_column] = pd.to_datetime(df[year_column], format='%Y')

    # Create right deflator

    if method == 'gdp_xe':
        deflator = utils.clean_df(utils.gdp_xe_deflator(base=base),
                                  country_column, year_column)

    elif method == 'gdp_ratio':
        deflator = utils.clean_df(utils.gdp_factor(output_type='usd', output_base=base),
                                  country_column, year_column)

    # Add deflator to provided data
    df = df.merge(deflator, on=[country_column, year_column], how='left')

    # Deflate values
    if to_current:
        df['value_current'] = df[value_column]*df['deflator']

    else:
        df['value_constant'] = df[value_column]/(df['deflator'])

    # drop deflator column
    df.drop('deflator', axis=1, inplace=True)

    # convert date column back to int if needed
    try:
        if date_int:
            df[year_column] = df[year_column].dt.year
    except:
        pass

    return df.reset_index(drop=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 21 13:07:16 2021

@author: jorge
"""

# %% Import required packages and create paths
import pandas as pd
import wbgapi as wb
import os


path = os.path.dirname(__file__)
ppath = os.path.dirname(path)


# %% update data

def update_exchange(test=False):

    try:

        e = wb.data.DataFrame('PA.NUS.FCRF', mrv=25, numericTimeKeys=True, labels=False,
                              columns='series', timeColumns=True).reset_index()

    except:
        raise ConnectionError(
            'Failed to retreive data from WorldBank. Try again or use update=False')

    e = e.rename(columns={'economy': 'iso_code',
                          'PA.NUS.FCRF': 'value',
                          'time': 'year'})

    e = e.dropna(subset=['value'])

    e.to_csv(path+r'/data/exchange_data.csv', index=False)
    print('Updated exchange data')
    if test:
        return e


def update_gdp(test=False):
    try:
        gdp_lcu = wb.data.DataFrame('NY.GDP.MKTP.CN', mrv=25, numericTimeKeys=True, labels=False,
                                    columns='series', timeColumns=True).reset_index()

        gdp_const = wb.data.DataFrame('NY.GDP.MKTP.KD', mrv=25, numericTimeKeys=True, labels=False,
                                      columns='series', timeColumns=True).reset_index()

    except:
        raise ConnectionError(
            'Failed to retreive data from WorldBank. Try again or use update=False')

    gdp_lcu = gdp_lcu.rename(columns={'economy': 'iso_code',
                                      'NY.GDP.MKTP.CN': 'value',
                                      'time': 'year'})

    gdp_const = gdp_const.rename(columns={'economy': 'iso_code',
                                          'NY.GDP.MKTP.KD': 'value',
                                          'time': 'year'})

    gdp_lcu.to_csv(path+r'/data/gdp_lcu_raw.csv', index=False)
    gdp_const.to_csv(path+r'/data/gdp_const_raw.csv', index=False)

    print('Updated GDP data')
    if test:
        return gdp_lcu, gdp_const


def update_data(indicators=['xe', 'gdp']):

    # Exchange rate data
    indicators_dict = {'xe': update_exchange, 'gdp': update_gdp}

    if isinstance(indicators, str):
        if indicators == 'all':
            _, _ = update_exchange(), update_gdp()
        else:
            raise ValueError(f'"{indicators}" is not valid. Indicators must be a list of valid indicators, or \
                             the string "all" to update all indicators at once')

    elif isinstance(indicators, list):

        for i in indicators:
            if i not in indicators_dict.keys():
                raise ValueError(f'Invalid indicator {i}')
            indicators_dict[i]()

    else:
        raise TypeError(
            f'Indicators list is of an invalid type "{type(indicators)}"')


# %%

def clean_df(df, country_column, year_column):
    """Rename/clean df to match provided data"""

    df.columns = [country_column, year_column, 'deflator']

    if df[year_column].dtype != 'datetime64[ns]':
        df[year_column] = pd.to_datetime(df[year_column], format='%Y')

    return df


def exchange_rates():
    """
    Gets exchange rates from the World Bank (indicator PA.NUS.FCRF) for the last
    25 non-empty values for all countries.


    Returns
    -------
    A Pandas DataFrame
        Containing 3 columns: iso_code, year, and value (the exchange rates).

    """

    return pd.read_csv(path+r'/data/exchange_data.csv')


def country_rebase(df: pd.DataFrame, c: str, base: int = 2019,
                   value_column='value', year_column='year') -> pd.DataFrame:
    """ Function used to rebase deflators or exchange rates.

        It should be applied to a dataframe, and the country name must be 
        passed individually as a string.

        The base year can be optionall defined (2019 is default)"""

    # Error handling
    if type(base) != int:
        raise ValueError('The base year must be passed as an integer')

    if type(df) != pd.DataFrame:
        raise TypeError('A DataFrame must be provided')

    if df[year_column].dtype == 'datetime64[ns]':
        df[year_column] = df[year_column].dt.year
        year_is_date = True

    # Filter the dataframe to keep only the selected country
    country = df.loc[df.iso_code == c].reset_index(drop=True)

    # Convert the year column to integer
    country[year_column] = country[year_column].astype(int)

    # Try to create a variable containing the value for the base year. If base
    # Year is unavailable, an error will be printed but script will continue
    try:
        base = float(country.loc[country[year_column]
                                 == base, value_column].values)
    except:
        pass

    country['deflator'] = 100*(country[value_column] / base)

    try:
        if year_is_date:
            country[year_column] = pd.to_datetime(
                country[year_column], format='%Y')
    except:
        pass

    return country[['iso_code', year_column, 'deflator']]


def gdp_deflator(base: int = 2019) -> pd.DataFrame:
    """Return gdp deflators per country per year"""

    # Read the deflators data. Some cleaning done manually before
    df = pd.read_csv(path+r'/data/deflators_data.csv')

    # Melt dataframe
    df = df.melt(id_vars='iso_code', var_name='year')

    # Remove number formatting
    df.value = df.value.str.replace(',', '')

    # Change number type to float
    df.value = df.value.astype(float)

    # Check if selected base year is available in the data
    if base not in list(set(df.year.astype(int))):
        raise ValueError(
            f'The base year provided "{base}" is not available in the data')

    # Create a country name set
    countries = list(set(df.iso_code))

    # Return dataframe with all countries rebased
    return pd.concat([country_rebase(df, c, base=base) for c in countries], ignore_index=True)


def exchange_deflator(base: int = 2019) -> pd.DataFrame:
    """Return exchange deflators per country per year."""

    # Using the wb data api, get foreign exchange data
    e = exchange_rates()

    countries = list(set(e.iso_code))

    return pd.concat([country_rebase(e, c, base=base) for c in countries], ignore_index=True)


def gdp_xe_deflator(base: int = 2019):

    # Import and format gdp data
    gdp_def_data = gdp_deflator(base=base)

    # Import and format exchange data
    xe_def_data = exchange_deflator(base=base)

    # Create deflators dataframe
    dft = gdp_def_data.merge(xe_def_data, on=['iso_code', 'year'], how='inner',
                             suffixes=('_gdp', '_currency'))

    # Create combined deflator
    dft['deflator'] = dft.deflator_gdp/dft.deflator_currency

    return dft[['iso_code', 'year', 'deflator']].reset_index(drop=True)


def gdp_factor(year: int = 2010, output_type: str = 'usd', output_base=2010) -> pd.DataFrame:
    """
    Calculates a gdp deflator factor, sometimes used to convert LCU into constant USD.
    For example, ReSAKKS calculates this factor by dividing current LCU GDP by constant USD GDP.

    This script returns the LCU-USD Constant conversion factor, or the equivalent factor
    for converting current USD into Constant USD.

    Parameters
    ----------
    year : int, optional
        The base year used by the World Bank for its constant GDP numbers. The default is 2010.
    output_type : str, optional
        Selects whether output can be applied to convert constant usd to current usd ('usd')
        or to convert usd to current LCU ('lcu'). The default is 'usd'.
    update : bool, optional
        Whether to update the underlying WB GDP data. The default is False.

    Raises
    ------
    ValueError
        A value error is raised if the output type is incorrectly specified.

    Returns
    -------
    Pandas DataFrame
        The output of this function is a Pandas DataFrame containing only 3 
        columns: iso_code (str), year (datetime), and factor (float).

    """
    # Error handling
    if output_type not in ['usd', 'lcu']:
        raise ValueError(f'{output_type} must be "usd" or "lcu"')

    gdp_lcu = pd.read_csv(path+r'/data/gdp_lcu_raw.csv')
    gdp_const = pd.read_csv(
        path+r'/data/gdp_const_raw.csv')

    # Merge gdp in lcu and gdp in constant
    factor = gdp_lcu.merge(
        gdp_const, on=['iso_code', 'year'], suffixes=('_lcu', '_constant'))

    # Calculate factor
    factor['factor'] = factor.value_lcu/factor.value_constant

    if output_type == 'usd':
        # get exchange rates
        factor = factor.merge(exchange_rates(), on=[
                              'iso_code', 'year'], how='left')

        # correct exchange rates for ZWE which uses usd since 2009
        factor.loc[(factor.iso_code == 'ZWE') & (
            factor.year > 2008), 'value'] = 1

        # Produce usd factor. This means data in constant usd will be returned in current usd
        factor.factor = (factor.factor/factor.value)

    # Drop missing data
    factor = factor.dropna(subset=['factor'])

    countries = list(set(factor['iso_code']))

    factor = pd.concat([country_rebase(factor, c, base=output_base, value_column='factor') for c in countries],
                       ignore_index=True)

    factor.deflator = factor.deflator/100

    # Convert year to datetime
    factor.year = pd.to_datetime(factor.year, format='%Y')

    # keep only relevant variables
    return factor[['iso_code', 'year', 'deflator']].reset_index(drop=True)
